/*
 schedtool
 Copyright (C) 2002 Freek <freeky@uni.de>
 Release under GPL, version 2
 Use at your own risk.
 Inspired by setbatch (C) 2002 Ingo Molnar

 Born in the need of querying and setting SCHED_* policies.
 All output, even errors, go to STDOUT to ease piping.

 Setting CPU-affinity is now supported, too.

 schedtool can now be used to exec a new process with specific parameters.

 Content:

 main code
 cmd-line parsing
 the engine
 set_/print_process
 usage

 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sched.h>
#include <unistd.h>

#include "error.h"
#include "util.h"

/* provide support for the syscalls even if the libc doesn't know about it */
#ifdef HAVE_AFFINITY
#ifndef __NR_sched_setaffinity
#include "syscall_magic.h"
#endif
#endif

/* various operation modes: print/set/affinity/fork */
#define MODE_NOTHING	0x0
#define MODE_PRINT	0x1
#define MODE_SETPOLICY	0x2
#define MODE_AFFINITY	0x4
#define MODE_FORK	0x8

#define VERSION "0.99"

/*
 constants are from the O(1)-sched kernel's include/sched.h
 I don't want to include kernel-headers.
 Included those defines for improved readability.
 */
#undef SCHED_NORMAL
#undef SCHED_FIFO
#undef SCHED_RR
#undef SCHED_BATCH
#undef SCHED_OTHER
#define SCHED_NORMAL	0
#define SCHED_FIFO	1
#define SCHED_RR	2
#define SCHED_BATCH	3
#define SCHED_ISO	4
/* compatibility */
#define SCHED_OTHER	SCHED_NORMAL

#define CHECK_RANGE_POLICY(p) (p <= 4 && p >= 0)

char *TAB[]={"N: SCHED_NORMAL", "F: SCHED_FIFO", "R: SCHED_RR", "B: SCHED_BATCH", "I: SCHED_ISO", 0};

/* call it engine_s in lack of a better name */
struct engine_s {

	int mode;
	int policy;
	int prio;
	unsigned long aff_mask;

	/* # of args when going in PID-mode */
	int n;
	char **args;
};


void dump_table(void);
void usage(void);
int engine(struct engine_s *);
int set_process(pid_t, int, int);
int set_affinity(pid_t, unsigned long);
void print_process(pid_t);


extern char *optarg;
extern int optind, opterr, optopt;
extern int errno;

int main(int ac, char **dc)
{
	/* prio is 0 per default */
	int policy=-1, c, prio=0, mode=MODE_NOTHING;

	/* default aff_mask to 0xFFFF FFFF FFFF FFFF== all CPUs */
	unsigned long aff_mask=(0-1);

	if (ac < 2) {
		usage();
                /* return 0, because querying parent is a valid action */
		return(0);
	}

	while((c=getopt(ac, dc, "+01234NFRBIa:p:evh")) != -1) {

		switch(c) {
		case '0':
		case 'N':
			policy=SCHED_NORMAL;
			break;
		case '1':
		case 'F':
			policy=SCHED_FIFO;
			break;
		case '2':
		case 'R':
			policy=SCHED_RR;
			break;
		case '3':
		case 'B':
			policy=SCHED_BATCH;
			break;
		case '4':
		case 'I':
			policy=SCHED_ISO;
			break;
		case 'a':
#ifdef HAVE_AFFINITY
			mode |= MODE_AFFINITY;
			aff_mask=strtol(optarg, NULL, 16);
                        break;
#else
			printf("ERROR: CPU-affinity is not supported\n");
                        return(-1);
#endif
		case 'e':
			mode |= MODE_FORK;
			break;
		case 'p':
			prio=atoi(optarg);
			break;
		case 'v':
			mode |= MODE_PRINT;
                        break;
		case 'V':
		case 'h':
			usage();
                        return(0);
		default:
			//printf("ERROR: unknown switch\n");
			usage();
			return(1);
		}
	}

	/* parameter checking */
	if((policy==SCHED_NORMAL || policy==SCHED_BATCH || policy==SCHED_ISO) && prio) {
		/* for BATCH and NORMAL prio is ignored and must be 0 */
		decode_error("%s call may fail as static PRIO must be 0 or omitted",
			     TAB[policy]);

	} else if((policy==SCHED_FIFO || policy==SCHED_RR)) {
#define CHECK_RANGE_PRIO(p) (p <= prio_max && p >= prio_min)

		/* FIFO and RR - check min/max priority */
		int prio_max=sched_get_priority_max(policy);
		int prio_min=sched_get_priority_min(policy);

		if(! CHECK_RANGE_PRIO(prio)) {
			if(prio == 0) {
				decode_error("missing priority; specify static priority via -p");
			} else {
				decode_error("PRIO %d is out of range %d-%d for %s",
					     prio,
					     prio_min,
					     prio_max,
					     TAB[policy]
					    );
			}
			/* treat as all calls have failed */
			return(ac-optind);
		}
#undef CHECK_RANGE_PRIO
	}

        /* -1 means, no policy was set */
	if(policy != -1) {
		mode |= MODE_SETPOLICY;
	} else {
		mode |= MODE_PRINT;
	}

	if( mode_set(mode, MODE_FORK) && ! ( mode_set(mode, MODE_SETPOLICY)
					   || mode_set(mode, MODE_AFFINITY) )) {
		/* we have nothing to do */
		decode_error("Option -e needs scheduling-parameters, not given - exiting\n");
                return(-1);
	}


	{
                struct engine_s stuff;
		/* now fill struct */
		stuff.mode=mode;
		stuff.policy=policy;
		stuff.prio=prio;
		stuff.aff_mask=aff_mask;

                /* we have this much real args/PIDs to process */
		stuff.n=ac-optind;

                /* this is the first real arg */
		stuff.args=dc+optind;

		/* now go on and do what we were told */
		return(engine(&stuff));
	}
}


int engine(struct engine_s *e)
{
	int ret=0;
	int i;

#ifdef DEBUG_ME
	printf("Dumping mode: 0x%x\n", e->mode);
	printf("We have %d args to do\n", e->n);
	for(i=0;i < e->n; i++) {
		printf("Dump arg %d: %s\n", i, e->args[i]);
	}
#endif

	/* EXECUTE: handle -e seperately */
	if(mode_set(e->mode, MODE_FORK)) {

		char **new_argv=e->args;

		/*
		 at the moment, schedtool will set itself to the given params
		 then exec the new command, then exit itself.
                 Error-handling is a bit easier.
		 */
 		if(set_process(0, e->policy, e->prio)) {
                        return(-1);
		}

#ifdef HAVE_AFFINITY
		if(mode_set(e->mode, MODE_AFFINITY) && set_affinity(0, e->aff_mask)) {
			return(-1);
		}
#endif

#ifdef DEBUG_ME
		/* dump the newly built argv */
		{
			char **tmp=e->args;

			printf("Dumping new argv:\n");
			while(*tmp) {
				printf("%s\n", *tmp++);
			}
		}
#endif
		if(mode_set(e->mode, MODE_PRINT)) {
			print_process(getpid());
		}

		ret=execvp(*new_argv, new_argv);
		decode_error("schedtool: Could not exec %s", *new_argv);
		return(ret);
	}

	/*
	 handle normal query/set operation:
	 set/query all given PIDs
	 */
	for(i=0;i < e->n;i++) {

		int pid, tmpret=0;

		pid=atoi(e->args[i]);
		if(!pid) {
			decode_error("Ignoring arg %s: is not a PID", e->args[i]);
			continue;
		}

		if(mode_set(e->mode, MODE_SETPOLICY)) {
			/*
			 accumulate possible errors
			 the return value of main will indicate
			 how much set-calls went wrong
                         set_process returns -1 upon failure
			 */
			tmpret=set_process(pid,e->policy,e->prio);
			ret += tmpret;

                        /* don't proceed as something went wrong already */
			if(tmpret) {
				continue;
			}

		}
#ifdef HAVE_AFFINITY
		if(mode_set(e->mode, MODE_AFFINITY)) {
			tmpret=set_affinity(pid, e->aff_mask);

			if(tmpret) {
				continue;
			}

		}
#endif
		/* and print process info when set, too */
		if(mode_set(e->mode, MODE_PRINT)) {
			print_process(pid);
		}
	}

	/*
	 indicate how many errors we got; as ret is accumulated negative,
	 convert to positive
	 */
	return(abs(ret));
}


int set_process(pid_t pid, int policy, int prio)
{
	struct sched_param p;
	int ret;

	p.sched_priority=prio;

	/* anything other than 0 indicates error */
	if((ret=sched_setscheduler(pid, policy, &p))) {

		decode_error("could not set PID %d to %s",
			     pid,
			     (CHECK_RANGE_POLICY(policy) ? TAB[policy] : "unknown")
			    );

		return(ret);
	}
	return(0);
}


#ifdef HAVE_AFFINITY
int set_affinity(pid_t pid, unsigned long mask)
{
	int ret;

	if((ret=sched_setaffinity(pid, sizeof(mask), &mask))) {
		decode_error("could not set PID %d to affinity 0x%x",
			     pid,
                             mask
			    );
		return(ret);
	}
        return(0);
}
#endif


void print_process(pid_t pid)
{
	int policy;
	struct sched_param p;
        unsigned long aff_mask=(0-1);

	/* strict error checking not needed - it works or not. */
	if( ((policy=sched_getscheduler(pid)) < 0)
	    || (sched_getparam(pid, &p) < 0)
#ifdef HAVE_AFFINITY
	    || (sched_getaffinity(pid, sizeof(aff_mask), &aff_mask) < 0)
#endif
	  ) {

		decode_error("could not get scheduling-information for PID %d", pid);
	} else {
		printf("PID %5d: PRIO %3d, POLICY %-15s",
		       pid,
		       p.sched_priority,
		       TAB[policy]
		      );
#ifdef HAVE_AFFINITY
		printf(", AFFINITY 0x%lx", aff_mask);
#endif
                printf("\n");
	}
}



void usage(void)
{
	printf(
	       "get/set scheduling policies - v" VERSION ", GPL'd\n" \
	       "USAGE: schedtool PIDS                    - query PIDS\n" \
	       "       schedtool [OPTIONS] PIDS          - set PIDS\n" \
	       "       schedtool [OPTIONS] -e COMMAND    - exec COMMAND\n" \
               "\n" \
	       "set scheduling policies:\n" \
	       "    -N                    for SCHED_NORMAL\n" \
	       "    -F -p PRIO            for SCHED_FIFO       only as root\n" \
	       "    -R -p PRIO            for SCHED_RR         only as root\n" \
               "    -B                    for SCHED_BATCH\n" \
               "    -I                    for SCHED_ISO\n\n" \
	       "    -p STATIC_PRIORITY    usually 1-99; only for FIFO or RR\n" \
               "                          higher numbers means higher priority\n" \
	      );
#ifdef HAVE_AFFINITY
	printf("    -a AFFINITY_MASK      set CPU-affinity to bitmask (hex)\n\n");
#endif
	printf(
	       "    -e COMMAND [ARGS]     start COMMAND with specified policy/priority\n" \
               "    -v                    be verbose\n" \
	       "\n" \
	      );
/*	printf("Parent ");
	print_process(getppid()); */

}


void dump_table(void)
{
	int i;

	for(i=0;i<=3;i++) {
		printf("\t%d or %s\n", i, TAB[i]);
	}
}
